#ifndef defs_h__
#define defs_h__

#include "vecmat.h"
#include <vector>
using std::vector;

const float MAX_T = 1.0e30f;
const float TINY = 0.001f;

typedef vec3f RGBColor;
class Scene;
class Ray;
class Surface;

//==============================================================================
class Object
//==============================================================================
{
public:
  Object() {};
  virtual ~Object() {};

  virtual bool intersect( Ray& ray ) = 0;
  virtual RGBColor shade( const Ray& ray, Scene* scene ) = 0;
};


//==============================================================================
class Light
//==============================================================================
{
public:
  enum LightType
  {
    AMBIENT = 0,
    DIRECTIONAL = 1,
    POINT = 2
  };

  LightType type;
  vec3f v;
  RGBColor color;

  Light( LightType type, const RGBColor& color, const vec3f& v )
    : type(type), color(color), v(v)
  {
    if (type == DIRECTIONAL)
      this->v.normalize();
  }
};


//==============================================================================
class Surface
//==============================================================================
{
public:

  RGBColor baseColor; // base color of the surface
  float ka, kd, ks;   // ambient, diffuse, specular coefficients
  float ns;           // Shineness power
  float kr;           // reflection coef.
  float kt;           // transparency coef.
  float ior;          // index of refraction


  Surface( const RGBColor& baseColor, 
    float ka, float kd, float ks, float ns, 
    float kr, float kt, float ior) :
    baseColor( baseColor ), ka(ka), kd(kd), ks(ks), ns(ns),
    kr(kr), kt(kt), ior(ior)
  {}

  virtual RGBColor shade( const vec3f& P, const vec3f& N, const Ray& ray, 
    Scene* scene );

};

//==============================================================================
class Scene
//==============================================================================
{
public:
  vector<Object*> objects;
  vector<Light*> lights;
  vector<Surface*> surfaces;
  RGBColor backgroundColor;

  Scene() {}
  ~Scene();
};


//==============================================================================
class Ray
//==============================================================================
{
public:  
  vec3f origin;
  vec3f dir;
  int   depth;
  float t;
  Object* obj;

  Ray( const vec3f& origin, const vec3f& dir, int depth = 0 ) 
    : origin(origin), dir(dir), depth(depth), t(MAX_T), obj(NULL)
  {
    this->dir.normalize();
  }

  bool trace( Scene* scene )
  {
    for( int i=0; i < (int)scene->objects.size(); i++ )
      scene->objects[i]->intersect(*this);
    return obj != NULL;
  }

  RGBColor shade( Scene* scene )
  {
    return obj->shade(*this, scene);
  }
};

//==============================================================================
class RayCamera
//==============================================================================
{
public:
  vec3f eye;
  vec3f lookat;
  vec3f up;
  float fov;
  vec3f Du;
  vec3f Dv;
  vec3f Vp;
  
  RayCamera()
    : eye(0,0,0), lookat( 0,0,-1 ), up(0,1,0), fov(60)
  {
    init();
  }

  void init()
  {
    vec3f look = lookat - eye;
    Du = normalize(look.cross(up));
    Dv = normalize(look.cross(Du));
    float fl = float(1.0 / (2*tan((0.5*fov)*M_PI/180)));
    Vp = normalize(look);
    Vp = fl*Vp - 0.5f*( Du + Dv );
  }

  RGBColor renderPixel( float u, float v, Scene* scene )
  {
    vec3f dir = Vp + u*Du + v*Dv;
    Ray ray = Ray(eye, dir, 0);
    if (ray.trace(scene))
      return ray.shade(scene);
    else
      return scene->backgroundColor;
  }
};



//------------------------------------------------------------------------------
inline 
Scene::~Scene() 
{
  int i;
  for( i=0; i < (int)objects.size(); i++ )
    delete objects[i];
  for( i=0; i < (int)lights.size(); i++ )
    delete lights[i];
  for( i=0; i < (int)surfaces.size(); i++ )
    delete surfaces[i];
}



void Parse( const char* filename );



extern Scene* scene;
extern RayCamera* camera;
extern unsigned char* imageData;

#endif // defs_h__