#include "defs.h"
#include "sphere.h"

int lineNumber;
bool parseError;
char wspace[] = " \t\n";

//------------------------------------------------------------------------------
vec3f readVec3()
{
  vec3f ret;
  char* token;
  if( !parseError ) 
  {
    for( int i=0; i < 3; i++ )
    {
      token = strtok( NULL, wspace );
      if( !token )
      {
        printf("Expected number on line %d\n", lineNumber );
        parseError = true;
        break;
      }
      ret[i] = (float)atof( token );
    }
  } 
  return ret;
}

//------------------------------------------------------------------------------
float readFloat()
{
  float ret = 0;
  char* token;
  if( !parseError ) 
  {
    token = strtok( NULL, wspace );
    if( !token )
    {
      printf("Expected number on line %d\n", lineNumber );
      parseError = true;
    }
    else
      ret = (float)atof( token );
  } 
  return ret;
}

//------------------------------------------------------------------------------
void Parse( const char* filename )
{
  FILE* in = fopen( filename, "r" );
  if( in == NULL )
  {
    printf("Could not open file %s\n", filename );
    return;
  }

  // create  a new scene
  delete scene;
  scene = new Scene();

  // create a new camera
  delete camera;
  camera = new RayCamera();
  
  // create the default surface
  scene->surfaces.push_back( new Surface( vec3f(1,1,1), 0.2f, 0.6f, 0.6f, 30, 0, 0, 1 ) );
  Surface* currentSurface = scene->surfaces[0];

  // parse the file
  const int BUFSIZE = 256;
  char buffer[BUFSIZE];
  char *token;
  parseError = false;
  lineNumber = 0;
  while ( !feof(in) && !parseError ) 
  {
    if( !fgets( buffer, BUFSIZE, in ) )
      continue;
    lineNumber++;
    token = strtok( buffer, wspace);

    if( token == NULL || token[0] == '#' )
      continue;
    else if( strcmp( token, "sphere" ) == 0 )
    {
      vec3f center = readVec3();
      float radius = readFloat();
      scene->objects.push_back( new Sphere( center, radius, currentSurface ) );
    }
    else if( strcmp( token, "light" ) == 0 )
    {
      vec3f color = readVec3();
      token = strtok( NULL, wspace );
      if( token == NULL )
      {
        printf("Expected light type line %d\n", lineNumber );
        parseError = true;
      }
      else if( strcmp( token, "ambient" ) == 0 )
      {
        scene->lights.push_back( new Light( Light::AMBIENT, color, vec3f(0,0,0) ) );
      }
      else if( strcmp( token, "directional" ) == 0 )
      {
        vec3f dir = readVec3();
        scene->lights.push_back( new Light( Light::DIRECTIONAL, color, dir ) );
      }
      else if( strcmp( token, "point" ) == 0 )
      {
        vec3f pos = readVec3();
        scene->lights.push_back( new Light( Light::POINT, color, pos ) );   
      }
      else
      {
        printf( "Bad light type %s on line %d\n", token, lineNumber );
        parseError = true;
      }
    }
    else if( strcmp( token, "eye" ) == 0 )
    {
      camera->eye = readVec3();
    }
    else if( strcmp( token, "lookat" ) == 0 )
    {
      camera->lookat = readVec3();
    }
    else if( strcmp( token, "up" ) == 0 )
    {
      camera->up = readVec3();
    }
    else if( strcmp( token, "fov" ) == 0 )
    {
      camera->fov = readFloat();
    }
    else if( strcmp( token, "background" ) == 0 )
    {
      scene->backgroundColor = readVec3();
    }
    else if( strcmp( token, "surface" ) == 0 )
    {
      vec3f baseColor = readVec3();
      float ka = readFloat();
      float kd = readFloat();
      float ks = readFloat();
      float ns = readFloat();
      float kr = readFloat();
      float kt = readFloat();
      float ior = readFloat();
      scene->surfaces.push_back( 
        new Surface( baseColor, ka, kd, ks, ns, kr, kt, ior ) );
      currentSurface = scene->surfaces.back();
    }
    else
    {
      printf( "Bad token %s on line %d\n", token, lineNumber );
      parseError = true;
    }
  }

  printf( "primitives = %d   lights = %d \n", (int)scene->objects.size(),
    (int)scene->lights.size() );
  camera->init();
}
